import attr

from .core import FakerField


@attr.s(kw_only=True)
class NameField(FakerField):
    """
    Base for a name field which uses faker to generate. The supplied 'gender_field' points to the field in the fieldset
    which generates the gender value. It uses this field and row value to determine whether the generated name is male
    or female.
    """
    gender_field = attr.ib()

    def init_from_fieldset(self, fieldset):
        self.gender = fieldset.field_map.get(self.gender_field)

    def _next_value(self, row):
        if row.get(self.gender_field) == self.gender.male_value:
            return self._male_name()
        if row.get(self.gender_field) == self.gender.female_value:
            return self._female_name()


@attr.s(kw_only=True)
class TimeField(FakerField):
    """
    Create mock time using faker
    """
    format = attr.ib("%H:%M")

    def _next_value(self, row):
        return self._fake.time(pattern=self.format)


class FirstNameField(NameField):
    """
    Generate first name based on gender using the faker module (see NameField).
    """

    def _male_name(self):
        return self._fake.first_name_male()

    def _female_name(self):
        return self._fake.first_name_female()


class LastNameField(NameField):
    """
    Generate last name using the faker module.
    """

    def _male_name(self):
        return self._fake.last_name_male()

    def _female_name(self):
        return self._fake.last_name_female()


@attr.s(kw_only=True)
class MiddleNameField(NameField):
    """
    Mock middle name field.
    """

    first_name_field = attr.ib()

    def _male_name(self):
        return self._fake.first_name_male()

    def _female_name(self):
        return self._fake.first_name_female()

    def next_value(self, row):
        val = super().next_value(row)
        if val == "":
            return val
        if val == row.get(self.first_name_field):
            return self.next_value(row)

        return val


@attr.s
class AddressField(FakerField):
    """
    Mock address line field.
    """

    line_no = attr.ib()

    def _next_value(self, row):
        if self.line_no == 1:
            return self._fake.street_address()

        if self.line_no == 2:
            return self._fake.secondary_address()

        if self.line_no == 3:
            return self._fake.city()

        if self.line_no == 4:
            return ""


@attr.s(kw_only=True)
class PostcodeField(FakerField):
    """
    Mock postcode field.
    """

    def _next_value(self, row):
        return self._fake.postcode()


@attr.s(kw_only=True)
class PhoneField(FakerField):
    """
    Mock phone number field.
    """

    type = attr.ib(default='default')

    def _next_value(self, row):
        if self.type in ['cell', 'mobile']:
            return self._fake.cellphone_number()
        else:
            return self._fake.phone_number()


@attr.s(kw_only=True)
class EmailField(FakerField):
    """
    Create mock email using faker
    """
    safe = attr.ib(True)

    def _next_value(self, row):
        if self.safe:
            return self._fake.safe_email()
        else:
            return self._fake.email()


@attr.s(kw_only=True)
class PasswordField(FakerField):
    """
    Create mock password using faker
    """
    length = attr.ib(16)
    special_chars = attr.ib(True)
    digits = attr.ib(True)
    upper_case = attr.ib(True)
    lower_case = attr.ib(True)

    def _next_value(self, row):
        return self._fake.password(
            length=self.length,
            special_chars=self.special_chars,
            digits=self.digits,
            upper_case=self.upper_case,
            lower_case=self.lower_case
        )


@attr.s(kw_only=True)
class TextField(FakerField):
    """
    Generate  text using the faker.lorem provider. 'max_length' controls how long the text is allowed to be.
    """
    max_length: int = attr.ib(default=50)

    def _next_value(self, row):
        return self._fake.text(max_nb_chars=self.max_length)


@attr.s(kw_only=True)
class MemoField(FakerField):
    """
    Generate paragraphs using the faker.lorem provider. 'sentences' controls the exact number of sentences, when 'exact' is
    set to True, when set to False it will return a variable number of sentences (never less than 1) +- 40% of the
    'sentences' parameter.
    """
    sentences: int = attr.ib(default=3)
    exact: bool = attr.ib(default=False)

    def _next_value(self, row):
        return self._fake.paragraph(nb_sentences=self.sentences,
                                    variable_nb_sentences=not self.exact)



import random as rnd
import re
import uuid

import attr

from .error import ChangeValue

@attr.s
class Transformer:
    """
    Logic for transforming field row values. For example, changing string case, randomising insertion of errors.
    Multiple transformers can be specified for each field.
    """
    name = attr.ib(default=uuid.uuid4())

    def before_next(self, field, row):
        """
        Transformation before the value is obtained.

        Throw a headfake.error.ChangeValue exception with a new value if you want the value to be changed up the stack.
        :param field:
        :param row:
        :return:
        """


    def after_next(self, field, row, value):
        return value


class UpperCase(Transformer):
    """
    Converts value to upper case.
    """

    def after_next(self, field, row, value):
        return str(value).upper()


@attr.s(kw_only=True)
class IntermittentBlanks(Transformer):
    """
    Adds intermittent blank_values (default="") randomly at a rate specified by the blank_probability property.
    """
    blank_probability = attr.ib()
    blank_value = attr.ib(default="")

    def before_next(self, field, row):
        if rnd.random() < self.blank_probability:
            raise ChangeValue(self.blank_value)


@attr.s(kw_only=True)
class RegexSubstitute(Transformer):
    """
    Perform a regular expression substitution
    """
    pattern = attr.ib()
    replace = attr.ib()

    def after_next(self, field, row, value):
        return re.sub(self.pattern, self.replace, value)


@attr.s(kw_only=True)
class Truncate(Transformer):
    """
    Truncate a value
    """
    length = attr.ib()

    def after_next(self, field, row, value):
        return value[:int(self.length)]


@attr.s(kw_only=True)
class Padding(Transformer):
    """
    Pad a value to be a given length with the specified fill character
    """
    length = attr.ib()
    fill = attr.ib()
    align = attr.ib('left')

    def after_next(self, field, row, value):
        methods = {
            'left': value.ljust,
            'right': value.rjust,
        }
        pad = methods.get(self.align)
        if pad:
            return pad(self.length, str(self.fill))
        else:
            return value


@attr.s(kw_only=True)
class SplitPiece(Transformer):
    separator = attr.ib() #string to separator on
    index = attr.ib() #index of separated string to return

    def after_next(self, field, row, value):
        pieces = value.split(self.separator)

        if len(pieces)<self.index:
            return ""

        return pieces[self.index]


from headfake import field, Fieldset, HeadFake
import pytest
from unittest import mock
import datetime

row = {}

import random

class mock_datetime:
    @classmethod
    def now(cls):
        return datetime.date(2020, 3, 24)

def test_OptionValueField_chooses_value_basedon_probability_from_distribution(monkeypatch):
    n = mock.Mock()
    mock_choices = {"M":0.2,"F":0.8}
    n.return_value = "M"
    monkeypatch.setattr("random.choice",n)

    dob = field.OptionValueField(probabilities = mock_choices)

    assert dob.next_value(row) == "M"
    n.assert_called_with(dob._option_picks)


def test_ConcatField_joins_multiple_fields_together(monkeypatch):
    n = mock.Mock()
    n.side_effect = [123456787,123456787,123456783]
    monkeypatch.setattr("random.randrange",n)

    fields  = [field.ConstantField(value="X"),field.ConstantField(value="Y"),field.ConstantField(value="Z")]
    concat = field.ConcatField(fields = fields, glue=" ")

    assert concat.next_value(row) == "X Y Z"


import operator

def test_IfElseField_handles_simple_setup():
    fset = Fieldset(fields={"gender": ""})

    gender_cond = field.Condition(
        field = "gender",
        operator = operator.eq,
        value= "M"
    )

    if_else = field.IfElseField(
        condition = gender_cond,
        true_value = "MR",
        false_value = "MRS"
    )

    if_else.init_from_fieldset(fset)

    assert if_else.next_value({"gender":"M"}) == "MR"

    assert if_else.next_value({"gender": "F"}) == "MRS"

def test_IfElseField_handles_nested_setup():
    fset = Fieldset(fields={"marital_status": "M"})

    gender_cond = field.Condition(
        field="gender",
        operator=operator.eq,
        value="M"
    )

    ms_test_cond = field.Condition(
        field = "marital_status",
        operator = operator.eq,
        value= "M"
    )

    ms_if_else = field.IfElseField(
        condition=ms_test_cond,
        true_value="MRS",
        false_value="MISS"
    )

    gender_if_else = field.IfElseField(
        condition = gender_cond,
        true_value = "MR",
        false_value = ms_if_else
    )

    gender_if_else.init_from_fieldset(fset)

    assert gender_if_else.next_value({"gender":"M","marital_status":"M"}) == "MR"
    assert gender_if_else.next_value({"gender": "M", "marital_status": "S"}) == "MR"

    assert gender_if_else.next_value({"gender":"F","marital_status": "S"}) == "MISS"
    assert gender_if_else.next_value({"gender": "F", "marital_status": "M"}) == "MRS"

def test_IfElseField_handles_non_if_else_logic():
    fset = Fieldset(fields={"marital_status": "M"})
    HeadFake.set_seed(5)
    gender_cond = field.Condition(
        field="gender",
        operator=operator.eq,
        value="M"
    )


    ms_test_cond = field.Condition(
        field = "marital_status",
        operator = operator.eq,
        value= "M"
    )


    female_ov = field.OptionValueField(
        probabilities = {
            "MISS": 0.7,
            "MS": 0.1,
            "DR": 0.1,
            "PROF": 0.1
        })

    ms_if_else = field.IfElseField(
        condition=ms_test_cond,
        true_value="MRS",
        false_value=female_ov
    )


    gender_if_else = field.IfElseField(
        condition = gender_cond,
        true_value = "MR",
        false_value = ms_if_else
    )

    gender_if_else.init_from_fieldset(fset)

    assert gender_if_else.next_value({"gender":"M","marital_status":"M"}) == "MR"
    assert gender_if_else.next_value({"gender": "M", "marital_status": "S"}) == "MR"
    assert gender_if_else.next_value({"gender":"F","marital_status": "S"}) == "PROF"
    assert gender_if_else.next_value({"gender": "F", "marital_status": "S"}) == "MISS"
    assert gender_if_else.next_value({"gender": "F", "marital_status": "S"}) == "MISS"

def test_IfElseField_handles_condition_as_dictionary():
    fset = Fieldset(fields={"marital_status": "M"})
    HeadFake.set_seed(5)
    gender_cond = {
        "field":"gender",
        "operator":operator.eq,
        "value":"M"
    }


    ms_test_cond = {
        "field": "marital_status",
        "operator":operator.eq,
        "value": "M"
    }


    female_ov = field.OptionValueField(
        probabilities = {
            "MISS": 0.7,
            "MS": 0.1,
            "DR": 0.1,
            "PROF": 0.1
        })

    ms_if_else = field.IfElseField(
        condition=ms_test_cond,
        true_value="MRS",
        false_value=female_ov
    )


    gender_if_else = field.IfElseField(
        condition = gender_cond,
        true_value = "MR",
        false_value = ms_if_else
    )

    gender_if_else.init_from_fieldset(fset)

    assert gender_if_else.next_value({"gender":"M","marital_status":"M"}) == "MR"
    assert gender_if_else.next_value({"gender": "M", "marital_status": "S"}) == "MR"
    assert gender_if_else.next_value({"gender":"F","marital_status": "S"}) == "PROF"
    assert gender_if_else.next_value({"gender": "F", "marital_status": "S"}) == "MISS"
    assert gender_if_else.next_value({"gender": "F", "marital_status": "S"}) == "MISS"

def test_RepeatField_generates_list_of_values():
    random.seed(124)
    gender_field = field.RepeatField(
        field = field.GenderField(male_value="M",female_value="F",male_probability=0.6),
        min_repeats=3,
        max_repeats=7
    )

    assert gender_field.next_value({}) == ['M', 'M', 'F', 'M', 'F']
    assert gender_field.next_value({}) == ['M', 'M', 'M']

def test_RepeatField_generates_joined_list_of_values():
    random.seed(124)
    gender_field = field.RepeatField(
        field = field.GenderField(male_value="M",female_value="F",male_probability=0.6),
        min_repeats=3,
        max_repeats=7,
        glue=","
    )

    assert gender_field.next_value({}) == "M,M,F,M,F"
    assert gender_field.next_value({}) == "M,M,M"

def test_RepeatField_nested_generates_multiple_joined_list_of_values():
    random.seed(124)
    gender_field = field.RepeatField(
        field = field.RepeatField(
            field = field.GenderField(male_value="M",female_value="F",male_probability=0.6),
            min_repeats=3,
            max_repeats=7,
            glue=","
        ),
        min_repeats=2,
        max_repeats=8,
        glue="|"
    )

    assert gender_field.next_value({}) == "M,F,M|M,M,F,F,M|F,M,M,M,M,M|M,F,M,M,F"
    assert gender_field.next_value({}) == "M,M,F|F,M,M|F,F,M,M,F,M|F,M,M,M,M,M|F,M,M|F,M,M"

MOCK_NORM_NUM1 = 3.131313

def test_NumberField_generates_random_number_with_no_formatting(monkeypatch):
    monkeypatch.setattr("scipy.stats.norm",mock_norm)
    number_field = field.NumberField(
        distribution = "scipy.stats.norm",
        mean=2,
        sd=0.5
    )
    assert number_field.next_value({}) == MOCK_NORM_NUM1

class mock_norm:
    def __init__(self, **kwargs):
        pass

    def rvs(self):
        return MOCK_NORM_NUM1

def test_NumberField_generates_random_number_with_specified_decimal_places(monkeypatch):
    monkeypatch.setattr("scipy.stats.norm",mock_norm)
    random.seed(125)
    number_field = field.NumberField(
        distribution = "scipy.stats.norm",
        mean=2,
        sd=0.5,
        dp=2
    )
    assert number_field.next_value({}) == 3.13


def test_BooleanField_returns_default_values_if_no_parameters(monkeypatch):
    random.seed(120)
    gender = field.BooleanField()

    assert gender.next_value(row) == 0
    assert gender.next_value(row) == 1



def test_BooleanField_uses_value_parameters_if_provided(monkeypatch):

    random.seed(120)
    gender = field.BooleanField(true_value=True,false_value=False)

    assert gender.next_value(row) is False
    assert gender.next_value(row) is True


def test_BooleanField_returns_value_based_on_true_probability_parameter(monkeypatch):

    gender = field.BooleanField(true_probability=0.3)

    monkeypatch.setattr("random.random", lambda: 0.29)

    assert gender.next_value(row) == 1

    monkeypatch.setattr("random.random", lambda: 0.31)

    assert gender.next_value(row) == 0


def test_NumberField_returns_value_within_range_of_two_value(monkeypatch):

    fset = Fieldset(fields=[
        field.NumberField(
            name="num",
            distribution="scipy.stats.norm",
            sd=3,
            mean=0,
            min=-2,
            max=2,
            dp=1
        )
    ])
    HeadFake.set_seed(1234)
    df = fset.generate_data(5)

    for idx, item in df.iterrows():
        assert item['num']>-2 and item['num']<2

def test_NumberField_returns_value_within_range_of_two_existing_fields(monkeypatch):

    fset = Fieldset(fields=[
        field.NumberField(
            name="floor",
            distribution="scipy.stats.norm",
            sd=1,
            mean=-5,
            min=-3,
            dp=0
        ),
        field.NumberField(
            name="ceil",
            distribution="scipy.stats.norm",
            sd=1,
            mean=5,
            max=3,
            dp=0
        ),
        field.NumberField(
            name="num",
            distribution="scipy.stats.norm",
            sd=3,
            mean=0,
            min="floor",
            max="ceil",
            dp=1
        )
    ])

    df = fset.generate_data(5)

    for idx, item in df.iterrows():
        assert item['num']>item['floor'] and item['num']<item['ceil']

def test_NumberField_returns_value_within_range_of_two_embedded_fields(monkeypatch):

    fset = Fieldset(fields=[
        field.NumberField(
            name="num",
            distribution="scipy.stats.norm",
            sd=3,
            mean=0,
            min=field.NumberField(
                name="floor",
                distribution="scipy.stats.norm",
                sd=1,
                mean=-5,
                min=-3,
                dp=0
            ),
            max=field.NumberField(
                name="ceil",
                distribution="scipy.stats.norm",
                sd=1,
                mean=5,
                max=3,
                dp=0
            ),
            dp=1
        )
    ])

    df = fset.generate_data(5)

    for idx, item in df.iterrows():
        assert item['num']>-3 and item['num']<3

def test_DateField_returns_value_within_range_of_two_values_with_specified_sd_in_days(monkeypatch):
    min = datetime.date(2020,1,1)
    max = datetime.date(2020,6,1)

    fset = Fieldset(fields=[
        field.DateField(
            name="date",
            distribution="scipy.stats.norm",
            sd=10,
            mean=datetime.date(2020,3,1),
            min=min,
            max=max,

        )
    ])

    df = fset.generate_data(5)

    for idx, item in df.iterrows():
        assert item['date']>min and item['date']<max

def test_DateField_returns_value_within_range_of_two_values_with_specified_sd_in_years(monkeypatch):
    min = datetime.date(1954,1,1)
    max = datetime.date(2020,6,1)

    fset = Fieldset(fields=[
        field.DateField(
            name="date",
            distribution="scipy.stats.norm",
            sd=5,
            mean=datetime.date(1970,3,1),
            min=min,
            max=max,
            use_years=True
        )
    ])

    df = fset.generate_data(5)

    for idx, item in df.iterrows():
        assert item['date']>min and item['date']<max


def test_DateField_returns_value_within_range_of_two_values_with_specified_sd_in_years(monkeypatch):
    min = datetime.date(1954,1,1)
    max = datetime.date(2020,6,1)

    fset = Fieldset(fields=[
        field.DateField(
            name="date",
            distribution="scipy.stats.norm",
            sd=5,
            mean=datetime.date(1970,3,1),
            min=min,
            max=max,
            use_years=True
        )
    ])

    df = fset.generate_data(5)

    for idx, item in df.iterrows():
        assert item['date']>min and item['date']<max

def test_OptionValueField_raises_error_when_sum_of_probabilities_is_not_one():
    probs = {"A":0.8,"B":0.02}

    with pytest.raises(ValueError):
        field.OptionValueField(probabilities = probs)

    probs = {"A": 0.8, "B": 2}

    with pytest.raises(ValueError):
        field.OptionValueField(probabilities=probs)


def test_OptionValueField_handles_small_probability_when_choosing_options():
    probs = {"A":0.8,"B":0.02,"C":0.0001,"D":0.1799}

    ovf = field.OptionValueField(probabilities = probs)
    HeadFake.set_seed(1234)

    assert len(ovf._option_picks) == 10000
    assert len(list(filter(lambda x: x == "A", ovf._option_picks))) == 8000
    assert len(list(filter(lambda x: x == "B", ovf._option_picks))) == 200
    assert len(list(filter(lambda x: x == "C", ovf._option_picks))) == 1
    assert len(list(filter(lambda x: x == "D", ovf._option_picks))) == 1799
    assert ["A","A","A","A","D"] == [ovf.next_value(row) for i in range(1,6) ]

def test_OptionValueField_handles_very_small_probability_with_warning_when_choosing_options(monkeypatch):
    from unittest.mock import Mock
    warn = Mock()
    monkeypatch.setattr("warnings.warn", warn)
    probs = {"A":0.8,"B":0.02,"C":0.000001,"D":0.179999}

    ovf = field.OptionValueField(probabilities = probs)
    HeadFake.set_seed(1234)

    assert len(ovf._option_picks) == 1000000
    assert len(list(filter(lambda x: x == "A", ovf._option_picks))) == 800000
    assert len(list(filter(lambda x: x == "B", ovf._option_picks))) == 20000
    assert len(list(filter(lambda x: x == "C", ovf._option_picks))) == 1
    assert len(list(filter(lambda x: x == "D", ovf._option_picks))) == 179999
    assert ["B","A","A","A","A"] == [ovf.next_value(row) for i in range(1,6) ]
    warn.assert_called_with("Options include probabilities of 1e-6. This requires the creation of 1e6 possible options")


MALE_NAME = "JEFF"
MALE_NAME2 = "BIFF"
FEMALE_NAME = "JENNIFER"
FEMALE_NAME2 = "JANE"
MALE_SURNAME = "BRIDGES"
FEMALE_SURNAME = "SAUNDERS"
MALE_VALUE = "M"
FEMALE_VALUE = "F"
ADDRESS_LINE_1 = "15 THE DRIVE"
ADDRESS_LINE_2 = "SOMEWHERETON"
ADDRESS_LINE_3 = "SOMEWHERESHIRE"
ADDRESS_POSTCODE = "SW1 3AB"
PHONE_NUMBER = "01234 567890"


"""
Utility methods used by other parts of Headfake
"""
import re
import errno
import os

from importlib import import_module
from inspect import getsourcefile

from pathlib import Path

field_count = 0

def create_package_class(package_name):
    """
    Utility method which loads and returns the class corresponding to the provided package name.
    :param package_name: Fully qualified package name.
    :return:
    """
    if not isinstance(package_name, str):
        return package_name

    package_bits = package_name.split(".")
    class_name = package_bits.pop()
    module = import_module(".".join(package_bits))
    if not hasattr(module, class_name):
        raise TypeError(
            "Unable to setup the class/method '" + class_name + "' as it does not exist in the Python package '" + module.__name__ + "'")
    return getattr(module, class_name)


def create_class_tree(name, params):
    """Create recursive class tree given parameters.

    The name is only given if the original source had a key/value
    structure, otherwise None is supplied.

    Args:
        name:
            Name of the current class/object in the tree
        params:
            A dictionary or list of parameters used to build the class
        constant_cls:
            A class to wrap constant scalars (e.g. ints, strings, dates etc.)

    Returns:
        A class tree containing recursive classes, lists and dictionaries.

    """

    if isinstance(params, dict):
        class_name = params.get("class")
        if "class" in params:
            del params["class"]

        sub_params = {k: create_class_tree(k, v) for k, v in params.items()}

        if not class_name:
            return sub_params

        try:
            return create_package_class(class_name)(**sub_params)
        except TypeError as ex:
            handle_missing_keyword(ex)

    if isinstance(params, list):
        return [create_class_tree(p.get("name"), p) for p in params]

    return params

def calculate_age(start_date: "datetime.date", end_date: "datetime.date"):
    """
    Calculate age in years when given start date and end date
    :param start_date: Date object for start
    :param end_date: Date object for end
    :return:
    """
    try:
        birthday = start_date.replace(year=end_date.year)

        # raised when birth date is February 29
        # and the current year is not a leap year
    except ValueError:
        birthday = start_date.replace(year=start_date.year,
                                      month=start_date.month + 1, day=1)

    if birthday > end_date:
        return end_date.year - start_date.year - 1

    return end_date.year - start_date.year


def locate_file(file):
    """
    Locates a file either from the given path or in the package resources

    :param file: The file to locate. It can either be an absolute or relative path in the
    filesystem or in the package resources
    :return: a PosixPath of the file if found. If the file is not found then an exception is raised
    """

    path = Path(file)

    if not path.is_file():
        # if file does not exist then check to see if it's in the package resources

        py_module = import_module("headfake")
        path = Path(os.path.dirname(getsourcefile(py_module))) / path

    if path.is_file():
        return path.absolute()

    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(file))


def handle_missing_keyword(ex):
    """
    Used to provides more informative error when a keyword is missing from parameters.
    :param ex: Original exception
    :return:
    """
    kwonly_error = re.search("required keyword-only argument[s]{0,1}: ('.+')$", str(ex))

    if kwonly_error:
        raise TypeError(
            "The following required parameter(s) were missing: %s" %
            kwonly_error.group(1))

    unexpected_kw_error = re.search("got an unexpected keyword argument ('.+')$", str(ex))
    if unexpected_kw_error:
        raise TypeError(
            "The following unknown parameter was provided: %s" %
            unexpected_kw_error.group(1))
    raise ex

def new_field_name():
    global field_count
    field_count+=1
    return "field_" + str(field_count)


from headfake import field
from unittest import mock
import datetime

from tests.field.test_field_common import MALE_VALUE, FEMALE_VALUE

row = {}
import numpy as np

class mock_datetime:
    @classmethod
    def now(cls):
        return datetime.date(2020, 3, 24)


def test_GenderField_returns_male_value_if_random_no_is_lt_male_probability(monkeypatch):
    monkeypatch.setattr("random.random", lambda: 0.4999)
    gender = field.GenderField(male_value =MALE_VALUE, female_value =FEMALE_VALUE)
    obs = gender.next_value(row)
    assert obs == MALE_VALUE

def test_GenderField_returns_female_value_if_random_no_is_gt_or_eq_male_probability(monkeypatch):
    monkeypatch.setattr("random.random", lambda: 0.5)
    gender = field.GenderField(male_value =MALE_VALUE, female_value =FEMALE_VALUE)
    obs = gender.next_value(row)
    assert obs == FEMALE_VALUE

def test_DateOfBirthField_creates_date_using_value_from_distribution(monkeypatch):
    n = mock.Mock()

    monkeypatch.setattr("scipy.stats.norm",n)

    monkeypatch.setattr("datetime.datetime", mock_datetime)

    n.return_value.rvs.return_value=25.3
    dob = field.DateOfBirthField(distribution = "scipy.stats.norm", min=0, max=105, mean=45, sd=13, date_format="%d/%m/%Y")
    assert dob.next_value(row) == "06/12/1994"

def test_NhsNoField_generates_valid_nhs_number(monkeypatch):
    n = mock.Mock()
    n.return_value = 123456787
    monkeypatch.setattr("random.randrange",n)

    nhs_no = field.NhsNoField()

    assert nhs_no.next_value(row) == "123 456 7873"
    n.assert_called_with(100000000, 999999999)

def test_NhsNoField_handles_invalid_nhs_number_by_reselecting(monkeypatch):
    n = mock.Mock()
    n.side_effect = [123456789, 123456787]
    monkeypatch.setattr("random.randrange",n)

    nhs_no = field.NhsNoField()

    assert nhs_no.next_value(row) == "123 456 7873"

def test_NhsNoField_handles_duplicate_nhs_number_by_reselecting(monkeypatch):
    n = mock.Mock()
    n.side_effect = [123456787,123456787,123456783]
    monkeypatch.setattr("random.randrange",n)

    nhs_no = field.NhsNoField()

    assert nhs_no.next_value(row) == "123 456 7873"
    assert nhs_no.next_value(row) == "123 456 7830"

def test_AgeField_calculates_correct_age_using_to_and_from_fields_containing_dates():
    age = field.AgeField(from_value="from_date",to_value="to_date")
    assert age.next_value({"from_date":datetime.date(1953,3,5),"to_date":datetime.date(2010,5,4)}) == 57

def test_AgeField_calculates_correct_age_using_to_and_from_fields_containing_strings():
    age = field.AgeField(from_value="from_date",to_value="to_date", from_format="%d/%m/%Y",to_format="%d/%m/%Y")
    assert age.next_value({"from_date":"05/03/1953","to_date":"04/05/2010"}) == 57

def test_AgeField_calculates_correct_age_using_to_and_from_values_as_fields():
    np.random.seed(123)
    age = field.AgeField(
        from_value=field.DateField(
            min=datetime.date(1953,3,5),
            mean=datetime.date(1974,1,1),
            sd=13,
            max=datetime.date.today(),
            use_years=True,
            distribution="scipy.stats.norm"
        ),
        to_value=datetime.date.today()
    )
    assert age.next_value({}) == 61
    assert age.next_value({}) == 34
    assert age.next_value({}) == 43

def test_DeceasedField_simulates_death_based_on_risks_and_returns_additional_fields():
    from headfake.fieldset import Fieldset
    from headfake import HeadFake

    HeadFake.set_seed(543)
    dob = field.DateOfBirthField(distribution = "scipy.stats.norm", min=0, max=105, mean=45, sd=13, date_format="%d/%m/%Y")
    deceased = field.DeceasedField(
        deceased_true_value=1,
        deceased_false_value=0,
        dob_field="dob",
        deceased_date_field="dod",
        age_field="age",
        risk_of_death={"30-100":"2"},
        date_format = "%Y-%m-%d",
    )
    fieldset = Fieldset(fields={"dob":dob, "deceased":deceased})
    assert fieldset.generate_data(1).to_dict("records") == [{'age': 33, 'deceased': 1, 'dob': '01/02/1983', 'dod': '2017-12-05'}]


import random as rnd
import uuid
from abc import ABC, abstractmethod
from typing import List, Optional

import attr

from headfake.field import Field


@attr.s(kw_only=True)
class IdGenerator(ABC):
    """
    Base IdGenerator class which handles the generation of zero-padded values for ID fields. The key function is 'select_id'
    which returns a zero-filled ID number to the specified length.

    Args:
        length (int): Length of zero-padded ID value
        min_value (int): Minimum value/start point for ID value (default=1)
    """
    length: int = attr.ib()
    min_value: int = attr.ib(default=1)
    name: Optional[str] = attr.ib(default=uuid.uuid4())

    def select_id(self):
        val = self._select_number()
        return str(val).zfill(self.length)

    @abstractmethod
    def _select_number(self):
        """
        Generates the integer value used to create the ID in the 'select_id' function.
        :return:
        """
        pass


@attr.s(kw_only=True)
class IncrementIdGenerator(IdGenerator):
    """
    Incremental ID generator, increments by 1 each time an ID is generated.
    """
    current_no: int = attr.ib()

    @current_no.default
    def _default_current_no(self):
        return self.min_value

    def _select_number(self):
        if len(str(self.current_no)) > self.length:
            raise ValueError("next number is greater than length")

        val = self.current_no
        self.current_no += 1

        return val


@attr.s
class RandomIdGenerator(IdGenerator):
    """
    Base random ID generator. Sets up a maximum value.
    """
    max_value: int = attr.ib()

    @max_value.default
    def _default_max_value(self):
        return int("9" * self.length)


@attr.s
class RandomNoReuseIdGenerator(RandomIdGenerator):
    """
    Random unique ID generator which retains the used IDs so it does not reuse them.
    """
    _used_values: List[id] = attr.ib(factory=list)

    def _select_number(self):
        val = rnd.randrange(self.min_value, self.max_value)
        if val in self._used_values:
            return self._select_number()

        self._used_values.append(val)

        return val


@attr.s
class RandomReuseIdGenerator(RandomIdGenerator):
    """
    Random non-unique ID generator. Simply generates a random number between the min_value and max_value.
    """

    def _select_number(self):
        val = rnd.randrange(self.min_value, self.max_value)
        return str(val).zfill(self.length)


@attr.s(kw_only=True)
class IdField(Field):
    """
    Field which generates ID values using a provided IdGenerator class.

    Args:
        prefix: Prefix prepended to generated ID
        suffix: Suffix appended to generated ID
        generator: IdGenerator class name to use (defaults to [IncrementIdGenerator](#headfakefieldcoreincrementidgenerator)
    """
    prefix: str = attr.ib(default="")
    suffix: str = attr.ib(default="")
    generator: IdGenerator = attr.ib(default=IncrementIdGenerator(length=3))

    def _next_value(self, row):
        val = self.generator.select_id()

        return self.prefix + val + self.suffix


from headfake import field, Fieldset
from unittest import mock
import datetime

from tests.field.test_field_common import MALE_VALUE, FEMALE_VALUE, MALE_NAME, FEMALE_NAME, MALE_SURNAME, \
    FEMALE_SURNAME, MALE_NAME2, FEMALE_NAME2, ADDRESS_LINE_1, ADDRESS_LINE_2, ADDRESS_LINE_3, ADDRESS_POSTCODE, \
    PHONE_NUMBER

row = {}

import random

class mock_datetime:
    @classmethod
    def now(cls):
        return datetime.date(2020, 3, 24)



def test_FirstNameField_returns_male_name_if_gender_field_is_male(monkeypatch):
    f = mock.Mock()
    f.return_value.first_name_male.return_value=MALE_NAME

    monkeypatch.setattr("faker.Faker", f)

    fn_field = field.FirstNameField(gender_field="gender")

    gender = field.GenderField(male_value=MALE_VALUE, female_value=FEMALE_VALUE)

    fset = Fieldset(fields={"gender": gender})

    fn_field.init_from_fieldset(fset)

    obs = fn_field.next_value({"gender":MALE_VALUE})
    assert obs == MALE_NAME

def test_FirstNameField_returns_female_name_if_gender_field_is_female(monkeypatch):
    f = mock.Mock()
    f.return_value.first_name_female.return_value = FEMALE_NAME

    monkeypatch.setattr("faker.Faker", f)

    fn_field = field.FirstNameField(gender_field="gender")

    gender = field.GenderField(male_value=MALE_VALUE, female_value=FEMALE_VALUE)

    fset = Fieldset(fields={"gender": gender})

    fn_field.init_from_fieldset(fset)

    obs = fn_field.next_value({"gender": FEMALE_VALUE})
    assert obs == FEMALE_NAME

def test_LastNameField_returns_male_last_name_if_gender_is_male(monkeypatch):
    f = mock.Mock()
    f.return_value.last_name_male.return_value = MALE_SURNAME

    monkeypatch.setattr("faker.Faker", f)

    ln_field = field.LastNameField(gender_field="gender")

    gender = field.GenderField(male_value=MALE_VALUE, female_value=FEMALE_VALUE)

    fset = Fieldset(fields={"gender": gender})

    ln_field.init_from_fieldset(fset)

    obs = ln_field.next_value({"gender": MALE_VALUE})
    assert obs == MALE_SURNAME

def test_LastNameField_returns_female_last_name_if_gender_is_female(monkeypatch):
    f = mock.Mock()
    f.return_value.last_name_female.return_value = FEMALE_SURNAME

    monkeypatch.setattr("faker.Faker", f)

    ln_field = field.LastNameField(gender_field="gender")

    gender = field.GenderField(male_value=MALE_VALUE, female_value=FEMALE_VALUE)

    fset = Fieldset(fields={"gender": gender})

    ln_field.init_from_fieldset(fset)

    obs = ln_field.next_value({"gender": FEMALE_VALUE})
    assert obs == FEMALE_SURNAME

def test_MiddleNameField_returns_different_name_if_same_as_first_name(monkeypatch):
    f = mock.Mock()
    f.return_value.first_name_female.side_effect = [FEMALE_NAME, FEMALE_NAME2]

    monkeypatch.setattr("faker.Faker", f)
    fn_field = field.FirstNameField(gender_field="gender")

    mn_field = field.MiddleNameField(gender_field="gender", first_name_field="first_name")

    gender = field.GenderField(male_value=MALE_VALUE, female_value=FEMALE_VALUE)

    fset = Fieldset(fields={"first_name":fn_field,"gender": gender})

    mn_field.init_from_fieldset(fset)

    obs = mn_field.next_value({"gender": FEMALE_VALUE,"first_name":FEMALE_NAME})
    assert obs == FEMALE_NAME2

def test_MiddleNameField_returns_male_name_if_gender_is_male(monkeypatch):
    f = mock.Mock()
    f.return_value.first_name_male.return_value = MALE_NAME

    monkeypatch.setattr("faker.Faker", f)
    fn_field = field.FirstNameField(gender_field="gender")

    mn_field = field.MiddleNameField(gender_field="gender", first_name_field="first_name")

    gender = field.GenderField(male_value=MALE_VALUE, female_value=FEMALE_VALUE)

    fset = Fieldset(fields={"first_name":fn_field,"gender": gender})

    mn_field.init_from_fieldset(fset)

    obs = mn_field.next_value({"gender": MALE_VALUE,"first_name":MALE_NAME2})
    assert obs == MALE_NAME

def test_MiddleNameField_returns_female_name_if_gender_is_female(monkeypatch):
    f = mock.Mock()
    f.return_value.first_name_female.return_value = FEMALE_NAME

    monkeypatch.setattr("faker.Faker", f)
    fn_field = field.FirstNameField(gender_field="gender")

    mn_field = field.MiddleNameField(gender_field="gender", first_name_field="first_name")

    gender = field.GenderField(male_value=MALE_VALUE, female_value=FEMALE_VALUE)

    fset = Fieldset(fields={"first_name":fn_field,"gender": gender})

    mn_field.init_from_fieldset(fset)

    obs = mn_field.next_value({"gender": FEMALE_VALUE,"first_name":FEMALE_NAME2})
    assert obs == FEMALE_NAME

def test_DateOfBirthField_creates_date_using_value_from_distribution(monkeypatch):
    n = mock.Mock()

    monkeypatch.setattr("scipy.stats.norm",n)

    monkeypatch.setattr("datetime.datetime", mock_datetime)

    n.return_value.rvs.return_value=25.3
    dob = field.DateOfBirthField(distribution = "scipy.stats.norm", min=0, max=105, mean=45, sd=13, date_format="%d/%m/%Y")
    assert dob.next_value(row) == "06/12/1994"

def test_AddressField_outputs_address(monkeypatch):
    f = mock.Mock()
    f.return_value.street_address.return_value = ADDRESS_LINE_1
    f.return_value.secondary_address.return_value = ADDRESS_LINE_2
    f.return_value.city.return_value = ADDRESS_LINE_3

    monkeypatch.setattr("faker.Faker", f)

    assert field.AddressField(line_no=1).next_value(row) == ADDRESS_LINE_1
    assert field.AddressField(line_no=2).next_value(row) == ADDRESS_LINE_2
    assert field.AddressField(line_no=3).next_value(row) == ADDRESS_LINE_3

def test_PostcodeField_outputs_postcode(monkeypatch):
    f = mock.Mock()
    f.return_value.postcode.return_value = ADDRESS_POSTCODE

    monkeypatch.setattr("faker.Faker", f)

    assert field.PostcodeField().next_value(row) == ADDRESS_POSTCODE

def test_PhoneField_outputs_phone_number(monkeypatch):
    f = mock.Mock()
    f.return_value.phone_number.return_value = PHONE_NUMBER

    monkeypatch.setattr("faker.Faker", f)

    assert field.PhoneField().next_value(row) == PHONE_NUMBER

def test_TextField_outputs_random_text_shorter_than_the_default_50_chars():
    from faker import Faker
    Faker.seed(123)
    txt = field.TextField()

    assert txt.next_value({})=="Eaque quisquam eaque. Fugit natus exercitationem."

def test_TextField_outputs_random_text_shorter_than_max_length(monkeypatch):
    from faker import Faker
    Faker.seed(123)
    txt = field.TextField(max_length=10)

    assert txt.next_value({})=="Fugiat."

def test_MemoField_outputs_multiple_sentences(monkeypatch):
    from faker import Faker
    Faker.seed(123)
    mem = field.MemoField(sentences=2,exact=True)

    assert mem.next_value({})=="Eaque quisquam eaque. Fugit natus exercitationem."


"""
This file implements the HeadFake public API
"""

import random
import yaml
import numpy as np

from faker import Faker

from headfake.util import create_class_tree, locate_file


class HeadFake:
    """
    Provides the core logic as a class which has an input a parameter dictions.
    Includes support for populating the HeadFake class from a YAML file.
    """

    locale = "en_GB"
    field_count = 0
    def __init__(self, params, seed=None):
        """
        Creates an instance of the HeadFake object

        Args:
            params: parameters for generating data as a hierarchical dictionary
            seed: seed for initializing the pseudo-random generator
        """
        self.set_seed(seed)
        self.fieldset = self._create_fieldset(params)

    @staticmethod
    def from_yaml(filename, **kwargs):
        """
        Create an instance of the HeadFake class with parameters loaded from a .yaml file

        Args:
            filename: name of yaml template
            **kwargs: additional arguments passed to HeadFake constructor

        Returns:
            a HeadFake instance

        """
        path = locate_file(filename)
        with open(path) as file:
            params = yaml.safe_load(file)
        return HeadFake(params, **kwargs)

    @staticmethod
    def from_python(class_tree, **kwargs):
        """
        Create an instance of the HeadFake class from a pre-defined class tree

        Args:
            class_tree: the pre-defined class_tree
            **kwargs: additional arguments passed to HeadFake constructor

        Returns:
            a HeadFake instance

        Examples:
            ```python
            hf = HeadFake.from_python(
                headfake.Fieldset(
                    fields = {
                        "gender": headfake.field.GenderField(
                            male_value="M",
                            female_value="F",
                            male_probability=0.5
                        )
                ])
            )
            ```
        """

        return PyHeadFake(class_tree, **kwargs)


    @staticmethod
    def set_seed(seed):
        """
        Set the seed for initializing random number generator

        Args:
            seed: seed for initializing random number generator

        Returns:
            None

        """

        random.seed(seed)
        np.random.seed(seed)
        Faker.seed(seed)

    @classmethod
    def set_locale(cls, locale):
        """
        Set the locale for random value generation

        Args:
            locale: locale for generated data (e.g.

        Returns:
            None

        """
        cls.locale = locale

    def _create_fieldset(self, params):
        """
        Create the FieldSet from the parameters passed

        Args:
            params: parameters for the fieldset

        Returns:
            A FieldSet object
        """

        class_tree = create_class_tree(None, params)

        fieldset = class_tree.get("fieldset")


        return fieldset

    def generate(self, num_rows=1):
        """
        Generate fake data based on the parameters specified in the constructor

        Args:
            num_rows: number of rows to generate

        Returns:
            a pandas dataframe
        """

        return self.fieldset.generate_data(num_rows)

class PyHeadFake(HeadFake):
    def _create_fieldset(self, params):
        """
        Create the FieldSet from the parameters passed

        Args:
            params: parameters for the fieldset

        Returns:
            A FieldSet object
        """

        fieldset = params.get("fieldset")

        for field in fieldset.fields:
            field.init_from_fieldset(fieldset)

        return fieldset



"""
Fake/mock field generation logic
"""
__export__ = ["Field"]

import csv
import random as rnd
import uuid
from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional, Union
from datetime import datetime as dt, timedelta as td
import datetime

import attr
import faker

from headfake.error import ChangeValue
from headfake.transformer import Transformer
from headfake.util import create_package_class, locate_file, handle_missing_keyword, new_field_name

import numpy as np

@attr.s(kw_only=True)
class Field(ABC):
    """Acts as abstract Field class.

    Fields use the attrs module to handle available/required class attributes/defaults.

    Args:
        transformers (list): Optional list of transformer objects which act upon on field values at various points.
        name (str): Optional name of field (defaults to UUID4 string)

    """
    transformers: List[Transformer] = attr.ib(factory=list)
    name: Optional[str] = attr.ib(default=new_field_name())
    generate_after = False # force the value to be generated after other field values have been generated

    def after_init_params(self):
        [t.init_params(self) for t in self.transformers]

    def init_params(self):
        pass

    def init_from_fieldset(self, fieldset: "headfake.Fieldset"):
        """Initialises field in fieldset.

        This is run once all fields have been setup. It is generally used by fields in the fieldset to obtain
        information from other fields in the fieldset for later use.

        Args:
            fieldset: Fieldset to initialise from

        Returns:
            None
        """
        pass

    def next_value(self, row: Dict[str, Any]) -> Union[Any, Dict[str, Any]]:
        """Gets next generated value for field.

        Acts as a decorator around the private '_next_value' method.
        If 'transformers' have been provided in the constructor they can act on the value before or after it has been
        generated.

        Any headfake.error.ChangeValue exception thrown by a transformer will be caught and the supplied value will be
        returned.

        Args:
            row: The current data row as a dictionary

        Returns:
            Dictionary containing multiple fields OR a single field value

        """
        try:
            [t.before_next(self, row) for t in self.transformers]
        except ChangeValue as ex:
            return ex.value

        val = self._next_value(row)

        try:
            for t in self.transformers:
                val = t.after_next(self, row, val)
        except ChangeValue as ex:
            return ex.value

        return val

    @abstractmethod
    def _next_value(self, row: Dict[str, Any]):
        """Internal method that should be over-ridden in inheriting Field classes.

        It provides the generated next value.

        Args:
            row: The current data row as a dictionary

        Returns:
            Dictionary containing multiple fields OR a single field value
        """
        pass


@attr.s(kw_only=True)
class FakerField(Field):
    """Abstract base field for Faker-based value creation.
    """
    _fake = attr.ib()

    @_fake.default
    def _default_faker(self):
        from headfake import HeadFake
        fake = faker.Faker(HeadFake.locale)
        return fake


@attr.s(kw_only=True)
class OptionValueField(Field):
    """Field to generate option values, based on a provided dictionary of probabilities.

    Attributes:
        probabilities (dict): Dictionary of values/probabilities (e.g. {"A":0.2,"B":0.8})

    Raises:
        ValueError: when probabilities do not add up to 1
    """

    probabilities = attr.ib()
    _option_picks = attr.ib()

    @_option_picks.default
    def _default_option_picks(self):
        option_picks = []
        probs = [np.float32(prob) for prob in self.probabilities.values()]
        tot = np.sum(probs)

        if tot != 1:
            raise ValueError("Probabilities provided do not add up to 1")

        min_prob = min(self.probabilities.values())
        max_dp = count_decimal_places(min_prob)

        if max_dp>5:
            warnings.warn("Options include probabilities of 1e-" + str(max_dp) + ". This requires the creation of 1e" + str(max_dp) + " possible options")

        for val, prob in self.probabilities.items():
            num_picks = int((pow(10, max_dp)) * prob)
            option_picks += [val] * num_picks

        return option_picks

    def _next_value(self, row):
        return rnd.choice(self._option_picks)

import re
from locale import localeconv

import warnings

def count_decimal_places(value):
    value = str(value)
    dec_pt = localeconv()['decimal_point']
    decrgx = re.compile("\d+(%s\d+)?e(-)(\d+)" % dec_pt)
    find_e_value = decrgx.match(value)
    if find_e_value:
        return int(find_e_value.group(3))
    else:
        return len(value.split(dec_pt)[-1])

@attr.s(kw_only=True)
class DerivedField(Field):
    """Base field which uses an existing field type (_internal_field) to generate values.

    The class can be extended with new properties which can be used to setup the _internal_field.
    """
    @abstractmethod
    def _internal_field(self):
        pass

    def __attrs_post_init__(self):
        try:
            self._internal_field = self._internal_field()
        except TypeError as ex:
            handle_missing_keyword(ex)

    def _next_value(self, row):
        return self._internal_field.next_value(row)


@attr.s(kw_only=True)
class ConstantField(Field):
    """Field which generates constant values.

    Deprecated: can simply use scalar value (e.g. string, number)
    """
    value = attr.ib()

    def _next_value(self, row):
        return self.value


@attr.s(kw_only=True)
class ConcatField(Field):
    """
    Field which concatenates values generated by multiple fields together.

    You can also provide an optional glue string.
    """
    fields: List[Field] = attr.ib(factory=list)
    glue: str = attr.ib(default="")

    def _next_value(self, row):
        vals = []
        for field in self.fields:
            value = field.next_value(row)
            vals.append(value)

        return self.glue.join(vals)

    def init_from_fieldset(self, fieldset):
        for field in self.fields:
            field.init_from_fieldset(fieldset)


@attr.s(kw_only=True)
class MapFileField(Field):
    """
    Field which provides randomised file lookup based on a particular field.

    On initialisation this loads a CSV-based mapping file and randomises the rows based on a particular 'key_field'.
    It is then used in conjuction with one or more LookupMapFileFields to lookup values in other fields in the mapping
    file.
    """

    mapping_file = attr.ib()
    key_field = attr.ib()
    key_field_store = attr.ib()

    @key_field_store.default
    def _default_key_field_store(self):
        input_file = locate_file(self.mapping_file)
        with open(input_file, "r") as out:
            reader = csv.DictReader(out)
            data_map = {l.get(self.key_field): l for l in reader}

        return data_map

    def _next_value(self, row):
        rnd_key = rnd.choice(list(self.key_field_store.keys()))
        return rnd_key


@attr.s(kw_only=True)
class LookupMapFileField(Field):
    """
    Field which works with a specified MapFileField to get a data value from a specified column, using the key value from
    that field.
    """

    lookup_value_field = attr.ib()
    map_file_field = attr.ib()
    _map_file_field_obj = attr.ib(default=None)
    generate_after = True

    def _next_value(self, row):
        map_key = row.get(self._map_file_field_obj.name)
        map_line = self._map_file_field_obj.key_field_store.get(map_key)
        return map_line.get(self.lookup_value_field)

    def init_from_fieldset(self, fieldset):
        self._map_file_field_obj = fieldset.field_map.get(self.map_file_field)
        first_store_row = list(self._map_file_field_obj.key_field_store.values())[0]
        if self.lookup_value_field not in first_store_row:
            raise ValueError("Lookup value field  '%s' not found in file" % self.lookup_value_field)


@attr.s(kw_only=True)
class IfElseField(Field):
    """
    Field which generates a value based on the results of a condition.

    If the result is true, the true_value is generated/returned and if false, the false_value is generated.
    The true and false values can be field definitions or scalar values (e.g. strings, numbers, dates).


    """
    condition = attr.ib()
    true_value = attr.ib()
    false_value = attr.ib()
    _cond_obj = attr.ib()

    @_cond_obj.default
    def _default_cond_obj(self):
        return self.condition if isinstance(self.condition, Condition) else Condition(**self.condition)

    def _next_value(self, row):
        if self._cond_obj.is_true(row):
            return self.true_value.next_value(row) if hasattr(
                self.true_value, "next_value") else self.true_value
        else:
            return self.false_value.next_value(row) if hasattr(
                self.false_value, "next_value") else self.false_value


@attr.s(kw_only=True)
class Condition:
    """
    Dependent field condition which compares the results of a field expression with a value, using an operator function.

    In general it is easiest to use builtin Python "operator" functions such as `operator.or_`, `operator.and_`,
    `operator.eq` etc., however, it is possible to use a custom function if required (provided it accepts
    two values and returns a boolean value).
    """
    name = attr.ib(default=None)
    field = attr.ib()  # name of field to check or field definition (e.g. could nest IfElseField)
    operator = attr.ib()
    value = attr.ib()
    _operator_fn = attr.ib()

    @_operator_fn.default
    def _default_operator_fn(self):
        return create_package_class(self.operator)

    def is_true(self, row):
        return self._operator_fn(row.get(self.field), self.value)


@attr.s(kw_only=True)
class RepeatField(Field):
    """
    Field which repeats the generation of field values a random number of times within a specified range.

    The result is output as either a list or, if a glue string is provided, as a concatenated value separated by that
    string.
    """
    field: Field = attr.ib()
    min_repeats: int = attr.ib()
    max_repeats: int = attr.ib()
    glue: str = attr.ib(default=None)

    def _next_value(self, row):
        min_values = extract_number(self.min_repeats, row)
        max_values = extract_number(self.max_repeats, row)

        num_values = rnd.randrange(min_values, max_values)

        outputs = [self.field.next_value(row) for n in range(1, num_values + 1)]

        return self.glue.join(outputs) if self.glue else outputs


@attr.s(kw_only=True)
class NumberField(Field):
    """
    Field which generates a number based on a random float selection from a scipy statistical distribution.

    Available scipy distributions include all those from the scipy.stats module
    (see https://docs.scipy.org/doc/scipy/reference/stats.html).

    For custom distributions, any class can be used provided it follows the same constructor arguments (loc and scale)
    and has an rvs() function.

    Includes support for a mean and standard distribution. Also an optional decimal places (dp), min and max can be
    provided to produce numbers with these criteria.
    """
    distribution: str = attr.ib()
    mean: float = attr.ib()
    sd: float = attr.ib()
    min: float = attr.ib(default=None)
    max: float = attr.ib(default=None)
    dp: int = attr.ib(default=None)

    _dist_cls = attr.ib()

    @_dist_cls.default
    def _default_dist_cls(self):
        return create_package_class(self.distribution)(loc=self.mean, scale=self.sd)

    def _next_value(self, row):
        number = self._dist_cls.rvs()
        min = extract_number(self.min, row)
        max = extract_number(self.max, row)

        if (min and number < min) or (max and number > max):
            return self.next_value(row)

        if self.dp is not None:
            return round(number, self.dp)

        return number


@attr.s(kw_only=True)
class BooleanField(Field):
    """
    Field which generate a boolean (e.g. true/false) value based on a true_probability (default=0.5).

    The true_value (default=1) and false_value (default=0) which are returned can be specified.
    This field forms the basis of the derived GenderField
    """
    true_value: Any = attr.ib(default=1)
    false_value: Any = attr.ib(default=0)
    true_probability: float = attr.ib(default=0.5)

    def _next_value(self, row):
        return self.true_value if rnd.random() < self.true_probability else self.false_value


@attr.s(kw_only=True)
class DateField(NumberField):
    """
    Field which generates a date based on a random number of days from a scipy statistical distribution similarly to
    NumberField.

    The generated value is based on a mean date and standard deviation and the field 'adds' the generated number of
    days to the mean date.

    Available scipy distributions include all those from the scipy.stats module
    (see https://docs.scipy.org/doc/scipy/reference/stats.html).

    For custom distributions, any class can be used provided it follows the same constructor arguments (loc and scale)
    and has an rvs() function.

    Includes mean, min and max which can be provided as date objects or strings.
    If the latter, the mean_format, min_format and max_format define the date formats.
    """
    min_format: str = attr.ib(default=None)
    max_format: str = attr.ib(default=None)
    mean_format: str = attr.ib(default=None)
    format: str = attr.ib(default=None)
    use_years: bool = attr.ib(default=False)

    _dist_cls = attr.ib()

    @_dist_cls.default
    def _default_dist_cls(self):
        return create_package_class(self.distribution)(loc=0, scale=self.sd)

    def _next_value(self, row):
        min = extract_date(self.min, row, self.min_format)
        max = extract_date(self.max, row, self.max_format)
        mean = extract_date(self.mean, row, self.mean_format)

        num_to_mean = self._dist_cls.rvs()

        if self.use_years:
            num_to_mean *= 365.25

        date = mean + td(days=num_to_mean)

        if (min and date < min) or (max and date > max):
            return self.next_value(row)

        if self.format:
            return date.strftime(self.format)

        return date


def extract_number(value, row):
    """
    Extracts number to use for min/max or other input parameter.

    If value is a string, it will be assumed it is the name of a field in the row (and the value of that field will
    be returned).

    If it is a Field class, then the next_value function will be used to generate the value.

    Otherwise, the original value will be returned.

    :param value:
    :param row:
    :return:
    """
    if isinstance(value, str):
        if not value in row:
            raise ValueError(
                "Field with name '%s' is not present in the row or has not been generated yet" %
                value)
        return row.get(value)

    if isinstance(value, Field):
        return value.next_value(row)

    return value


def extract_date(value, row, date_format):
    """
    Extracts date to use for min/max or other input parameter.

    If value is a string, it will be assumed it is the name of a field in the row (and the value of that field will
    be converted into a date object.

    If it is a Field class, then the next_value function will be used to generate the value.

    Otherwise, the original value will be returned.

    :param value:
    :param row:
    :return:
    """
    if isinstance(value, str):
        if not value in row:
            raise ValueError(
                "Field with name '%s' is not present in the row or has not been generated yet" %
                value)

        data = row.get(value)
        if isinstance(data, datetime.date):
            return data

        return dt.strptime(data, date_format)

    if isinstance(value, Field):
        return value.next_value(row)

    return value



"""
This package contains the error classes raised by headfake.
"""


class ChangeValue(Exception):
    """
    Raised by a Transformer or Field when the generated value should change. The 'value' in the accessor is the new
    value.
    """

    def __init__(self, value):
        self.value = value
        super().__init__("Change value raised to " + str(self.value))



from headfake import HeadFake, Fieldset
from headfake.field import IdField


def test_Fieldset_can_accept_list_of_fields_as_input():
    data = {
        "fieldset":
            {
                "class": "headfake.fieldset.Fieldset",
                "fields": [
                    {"name":"spell_id",
                     "class":"headfake.field.IdField",
                     "prefix":"SPL",
                     "generator":{
                         "class":"headfake.field.IncrementIdGenerator",
                         "length":6,
                         "min_value":1000000
                     },
                    },
                    {"name":"admission_date",
                     "value":"2020-03-10"
                     },
                    {"name": "discharge_date",
                     "value": "2020-03-15"
                     },
                    {"name": "comments",
                     "class":"headfake.field.TextField",
                     "max_length":30
                     }
                ]
            }
    }

    hf = HeadFake(data)
    assert isinstance(hf.fieldset, Fieldset)
    assert isinstance(hf.fieldset.fields, list)
    assert list(hf.fieldset.field_map.keys()) == ["spell_id","admission_date","discharge_date","comments"]

def test_Fieldset_can_accept_constant_values_in_list_form():
    data = {
        "fieldset":
            {
                "class": "headfake.fieldset.Fieldset",
                "fields": [
                    {"name": "discharge_date",
                     "value": "2020-03-15"
                     }
                ]
            }
    }

    hf = HeadFake(data)


def test_Fieldset_can_accept_a_dictionary_of_fields_as_input():
    data = {
        "fieldset":
            {
                "class": "headfake.fieldset.Fieldset",
                "fields": {
                    "spell_id":{
                         "class":"headfake.field.IdField",
                         "prefix":"SPL",
                         "generator":{
                             "class":"headfake.field.IncrementIdGenerator",
                             "length":6,
                             "min_value":1000000
                         },

                    },
                    "admission_date": {
                         "class":"headfake.field.ConstantField",
                         "value":"2020-03-10"
                }
            }
            }
    }

    hf = HeadFake(data)
    assert isinstance(hf.fieldset, Fieldset)
    assert isinstance(hf.fieldset.fields, list)
    assert isinstance(hf.fieldset.field_map, dict)
    assert isinstance(hf.fieldset.field_map["spell_id"],IdField)
    assert hf.fieldset.field_map["spell_id"].name == "spell_id"



"""
This package includes fieldset classes
"""

import pandas as pd
from headfake.field import Field

class Fieldset:
    """
    The basic Fieldset object which contains the fields and parameters for the data generation process.
    """

    def __init__(self, fields, **kwargs):
        """
        constructor

        Args:
            fields: specification of fields in this fieldset
            **kwargs: dictionary of keyword arguments
        """

        if isinstance(fields, list):
            self.fields = fields
            self.field_map = {self._get_name(f): f for f in fields}
        else:
            for field_name, field in fields.items():
                if hasattr(field,"name") and field.name!= field_name:
                    field.name = field_name

            self.fields = list(fields.values())
            self.field_map = fields

        self.field_names = list(self.field_map.keys())

        for field in self.fields:
            if not hasattr(field, "init_from_fieldset"):
                continue

            field.init_from_fieldset(self)


    def _get_name(self, field):
        if hasattr(field,"name"):
            return field.name

        return field.get("name")

    def generate_data(self, num_rows):
        """
        Generates data based on the fields and parameters in this fieldset and return as a pandas dataframe

        Args:
            num_rows: number of rows to generate

        Returns:
            a pandas dataframe

        """
        dataset = pd.DataFrame(columns=self.field_names)

        generate_funcs = []
        after_funcs = []
        for field in self.fields:
            if field.generate_after is True:
                after_funcs.append((field.name, field.next_value))
            else:
                generate_funcs.append((field.name, field.next_value))

        generate_funcs.extend(after_funcs)

        for i in range(num_rows):
            row = {}
            for field_name, generate_fn in generate_funcs:
                next_val = generate_fn(row)
                if isinstance(next_val, dict):
                    row.update(next_val)
                else:
                    row[field_name] = next_val

            dataset = dataset.append(row, ignore_index=True)

        return dataset




import pytest

from headfake.util import create_class_tree
from headfake.field import TextField, IfElseField
from headfake.fieldset import Fieldset

def test_class_tree_from_field_dict_with_constant_is_created_correctly():

    tree = create_class_tree(None, {"discharge_date": "2020-03-15"})

    assert tree == {'discharge_date': "2020-03-15"}

def test_class_tree_from_field_list_with_constant_is_created_correctly():

    tree = create_class_tree(None, [{"name":"discharge_date", "value":"2020-03-15"}])

    assert tree == [{"name":"discharge_date", "value":"2020-03-15"}]

def test_nested_class_tree_with_constant_is_created_correctly():

    tree = create_class_tree(None, {"fieldset":{"fields":[{"name":"discharge_date", "value":"2020-03-15"},{"name":"comment","class":"headfake.field.TextField", "max_length":50}],"class":"headfake.fieldset.Fieldset"}})

    assert isinstance(tree.get("fieldset"),Fieldset)
    fields = tree.get("fieldset").fields

    assert fields[0] == {"name":"discharge_date", "value":"2020-03-15"}

    assert isinstance(fields[1],TextField)
    assert fields[1].name == "comment"
    assert fields[1].max_length == 50

def test_class_tree_with_dictionary_with_incorrect_class_raises_exception():
    with pytest.raises(TypeError):
        create_class_tree(None, {
            "title": {
                "class": "headfake.field.IfElseField",
                "condition": {
                    "field": "gender",
                    "operator": "operator.equal",
                    "value": "M"
                },
                "true_value": "MR",
                "false_value": "MRS"
            }
        })




import attr
import csv
import os
import tempfile

from headfake.cli import Command

@attr.s
class Args:
    template = attr.ib()
    output_file = attr.ib()
    no_rows = attr.ib()

    def as_list(self):
        return [
            "--output-file",
            self.output_file,
            "--no-rows",
            str(self.no_rows),
            self.template
        ]

def test_generate_patients_data_from_template():
    with tempfile.NamedTemporaryFile(mode="w",delete=False) as tmp:
        tmp.write(" ")
        tmp.close()
        args = Args(template = "examples/patients.yaml", no_rows = 20, output_file = str(tmp.name))

        Command.run(args.as_list())

        assert os.path.exists(tmp.name) is True

        with open(tmp.name,"r") as fh:
            file = csv.DictReader(fh)
            lines = list(file)

        assert len(lines) == 20

        os.unlink(tmp.name)


def test_generate_screening_data_from_template():
    with tempfile.NamedTemporaryFile(mode="w",delete=False) as tmp:
        tmp.write(" ")
        tmp.close()
        args = Args(template = "examples/screening.yaml", no_rows = 20, output_file = str(tmp.name))

        Command.run(args.as_list())

        assert os.path.exists(tmp.name) is True

        with open(tmp.name,"r") as fh:
            file = csv.DictReader(fh)
            lines = list(file)

        assert len(lines) == 20

        os.unlink(tmp.name)



import headfake.field.id
import pytest
from unittest import mock
import datetime

row = {}

class mock_datetime:
    @classmethod
    def now(cls):
        return datetime.date(2020, 3, 24)


def test_IncrementIdGenerator_returns_incremental_values():
    id = headfake.field.id.IncrementIdGenerator(length=2)

    assert ["01","02","03","04","05","06","07"]==[id.select_id() for i in range(1,8)]

def test_IncrementIdGenerator_fails_if_passes_maximum():
    id = headfake.field.id.IncrementIdGenerator(length=2)

    [id.select_id() for i in range(1, 100)]

    with pytest.raises(ValueError,match = r"next number is greater than length"):
        id.select_id()

def test_IdGenerator_is_correct_length():
    id1 = headfake.field.id.IncrementIdGenerator(length=6)
    assert id1.select_id() == "000001"
    assert id1.select_id() == "000002"

    id1 = headfake.field.id.IncrementIdGenerator(length=4)
    assert id1.select_id() == "0001"
    assert id1.select_id() == "0002"

def test_RandomNoReuseIdGenerator_generates_random_no_with_no_replacement(monkeypatch):
    id = headfake.field.id.RandomNoReuseIdGenerator(length=3)

    monkeypatch.setattr("random.randrange",mock.Mock(side_effect=[5,8,4,6,5,9]))
    assert id.select_id() == "005"
    assert id.select_id() == "008"
    assert id.select_id() == "004"
    assert id.select_id() == "006"
    assert id.select_id() == "009"

def test_RandomNoReuseIdGenerator_generates_random_no_with_replacement(monkeypatch):
    id = headfake.field.id.RandomReuseIdGenerator(length=3)

    monkeypatch.setattr("random.randrange",mock.Mock(side_effect=[5,8,4,6,5,9]))
    assert id.select_id() == "005"
    assert id.select_id() == "008"
    assert id.select_id() == "004"
    assert id.select_id() == "006"
    assert id.select_id() == "005"
    assert id.select_id() == "009"

def test_IdField_returns_values_from_id_field_type_with_suffix_and_prefix():
    id_generator = mock.MagicMock(headfake.field.id.IncrementIdGenerator)
    id_generator.select_id.side_effect = ["003","006","005","008","004"]

    id = headfake.field.id.IdField(prefix="P", suffix="S", generator=id_generator)

    assert id.next_value(row) == "P003S"
    assert id.next_value(row) == "P006S"
    assert id.next_value(row) == "P005S"
    assert id.next_value(row) == "P008S"
    assert id.next_value(row) == "P004S"



"""
This package implements the command line interface for headfake
"""

import argparse

from headfake import output, HeadFake


class Command:
    """
    Class for command line script
    """

    @staticmethod
    def run(args=None):
        """
        Entrypoint for command line script. this entry point is registered by setup.py when the
        package is installed.

        Returns:
            None
        """
        command = Command(args)
        command.execute()

    def __init__(self, args=None):
        """
        __init__ constructor.

        Initializes ArgumentParser and parses arguments passed at the command line.
        """

        parser = argparse.ArgumentParser(

            description="HEAlth Data Faker provides a command-line script to create mock data "
            "files based on a YAML-based template file (see examples/* for example "
            "templates). HeadFake uses the python package 'faker' to generate names and "
            "contact details.")

        parser.add_argument(
            "template",
            type=str,
            help="YAML-based template file describing structure of health data fields to generate"
        )

        parser.add_argument(
            "-o",
            "--output-file",
            help="Output data to file rather than STDOUT"
        )

        parser.add_argument(
            "-n",
            "--no-rows",
            type=int,
            help="Number of rows to generate",
            default=10
        )

        parser.add_argument(
            "-s",
            "--seed",
            type=int,
            help="Seed for the random data generator",
            required=False
        )

        self.args = parser.parse_args(args)

    def execute(self):
        """
        Runs the app by loading the configuration and generating data

        Returns:
            None
        """

        headfake = HeadFake.from_yaml(self.args.template, seed=self.args.seed)

        if self.args.output_file:
            outfile = output.CsvFileOutput(self.args)
        else:
            outfile = output.StdoutOutput(self.args)

        outfile.write(headfake.generate(self.args.no_rows))



#!/usr/bin/env python
import os
import sys
from itertools import chain

from setuptools import setup
from setuptools.config import read_configuration

# Append cwd for pip 19
sys.path.append(os.path.abspath("."))

################################################################################
# Programmatically generate some extras combos.
################################################################################
extras = read_configuration("setup.cfg")["options"]["extras_require"]

# Dev is everything
extras["dev"] = list(chain(*extras.values()))

# All is everything but tests and docs
exclude_keys = ("tests", "docs", "dev")
ex_extras = dict(filter(lambda i: i[0] not in exclude_keys, extras.items()))
# Concatenate all the values together for 'all'
extras["all"] = list(chain.from_iterable(ex_extras.values()))

setup(
    extras_require=extras,
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    package_data={'headfake': ['examples/*', 'examples/test_data/*']}
)



from abc import ABC, abstractmethod


class Output(ABC):
    """
    Base output class.
    """

    @abstractmethod
    def write(self, dataframe):
        """
        Main method used to write the fieldset to the output.
        :param dataframe:
        :return:
        """
        pass


class FileOutput(Output):
    """
    Output generated mock data for a single fieldset to the file specified in the options (output_file)
    """

    def __init__(self, options):
        """
        Setup output with appropriate options.
        :param options: Arguments dictionary
        """
        self.output_file = options.output_file

    @abstractmethod
    def write(self, dataframe):
        pass


class CsvFileOutput(Output):
    """
    Output generated mock data for a single fieldset to a CSV file specified in the
    options (output_file)
    """

    def __init__(self, options):
        """
        Setup output with appropriate options.
        :param options: Arguments dictionary
        """
        self.output_file = options.output_file

    def write(self, dataframe):
        dataframe.to_csv(self.output_file)


class JsonFileOutput(Output):
    """
    Output generated data for a single fieldset to a JSON text file specified in the options (output_file)
    """

    def __init__(self, options):
        """
        Setup output with appropriate options.
        :param options: Arguments dictionary
        """
        self.output_file = options.output_file

    def write(self, dataframe):
        dataframe.to_json(self.output_file)


class StdoutOutput(FileOutput):
    """
    Output generated data for a single fieldset to the console/STDOUT as CSV data
    """

    def write(self, dataframe):
        print(dataframe.to_csv(index=False))



from .core import Field, DerivedField, NumberField, BooleanField, extract_date
import attr
import datetime
import random as rnd
from typing import Dict, List, Any

from headfake.util import calculate_age


@attr.s(kw_only=True)
class DateOfBirthField(DerivedField):
    """
    Mock date of birth field. Calculates age based on random float selection from a scipy statistical distribution.
    This is multiplied by 365.25 to get age in days and a delta age (in days) from now is determined.
    The date is then output according to the date_format property.

    A min and max property are also required to keep the distribution within a particular range.

    """
    distribution: str = attr.ib()
    mean: float = attr.ib()
    sd: float = attr.ib()
    min: float = attr.ib()
    max: float = attr.ib()
    date_format: str = attr.ib()

    def _internal_field(self):
        return NumberField(mean=self.mean, sd=self.sd, min=self.min,
                           max=self.max, distribution=self.distribution)

    def _next_value(self, row):
        age_in_years = super()._next_value(row)

        age_in_days = age_in_years * 365.25
        dob = datetime.datetime.now() - datetime.timedelta(days=age_in_days)
        return dob.strftime(self.date_format)


@attr.s(kw_only=True)
class GenderField(DerivedField):
    """
    Field which generates gender values according to a 'male_probability' (default=0.5). Options include male_value (value of male selection), female_value (value of female
    selection and male_probability (probability that gender is male, default=0.5).

    e.g.
    ```python
    field = GenderField(male_value="M",female_value="F", male_probability=0.55)
    field.next_value(row)

    M
    F
    M
    M
    F
    ```
    """
    male_probability = attr.ib(default=0.5)
    male_value = attr.ib()
    female_value = attr.ib()

    def _internal_field(self):
        return BooleanField(true_value=self.male_value,
                            false_value=self.female_value, true_probability=self.male_probability)


@attr.s(kw_only=True)
class NhsNoField(Field):
    """
    Mock NHS number field which creates valid NHS numbers with the correct checksum digit.
    See https://www.closer.ac.uk/wp-content/uploads/CLOSER-NHS-ID-Resource-Report-Apr2018.pdf for details.
    """
    _used_values: List[id] = attr.ib(factory=list)

    def _next_value(self, row):
        val = rnd.randrange(100000000, 999999999)
        if val in self._used_values:
            return self._next_value(row)

        self._used_values.append(val)

        strval = str(val)
        checksum = 0

        multiplier = 10

        for char in strval:
            checksum += (multiplier * int(char))
            multiplier -= 1

        checkdigit = 11 - (checksum % 11)
        if checkdigit == 11:
            checkdigit = 0

        if checkdigit == 10:
            return self.next_value(row)

        return strval[0:3] + " " + strval[3:6] + " " + strval[6:9] + str(checkdigit)


@attr.s(kw_only=True)
class DeceasedField(Field):
    """
    Deceased mock field which uses a list of age range/mortality risk and a simulated patient 'aging' to determine if a
    patient is deceased and when they died. The ages are specified in a dictionary of the form {"X1-Y1":R1, "X2-Y2":R2}.

    The risk is defined as the 1 in R risk of death.

    The end date to use to determine the age defaults to today, but can be set using the 'end_date' argument as a date
    object, a field name or a Field object.

    """
    deceased_true_value = attr.ib(default=1)
    deceased_false_value = attr.ib(default=0)
    dob_field: str = attr.ib()
    deceased_date_field: str = attr.ib(default=None)

    age_field: str = attr.ib(default=None)

    risk_of_death: Dict[str, str] = attr.ib()
    date_format = attr.ib()
    _risk_by_age: Dict[int, float] = attr.ib()

    end_date = attr.ib(default=datetime.date.today())
    end_date_format = attr.ib(default=None)
    generate_after = True

    @_risk_by_age.default
    def _default_risk_by_age(self):
        risk_by_age = {}
        for k, v in self.risk_of_death.items():
            from_age, to_age = k.split("-")

            risk = 1 / int(v)
            for i in range(int(from_age), int(to_age) + 1):
                risk_by_age[i] = risk

        return risk_by_age

    def init_from_fieldset(self, fieldset):
        self._dob_field = fieldset.field_map.get(self.dob_field)

        if self.deceased_date_field:
            fieldset.field_names.append(self.deceased_date_field)

        if self.age_field:
            fieldset.field_names.append(self.age_field)



    def _next_value(self, row):
        dob = row.get(self.dob_field)
        dob = datetime.datetime.strptime(dob, self._dob_field.date_format).date()

        today = extract_date(self.end_date, row, self.end_date_format)
        prev_date = dob
        curr_date = dob + datetime.timedelta(weeks=52)

        while (curr_date < today):
            curr_age = calculate_age(dob, curr_date)
            curr_risk = self._risk_by_age.get(curr_age)

            if curr_risk is not None and rnd.random() <= curr_risk:
                int_bt_curr_and_prev = curr_date - prev_date
                rnd_days_after_curr = rnd.randrange(0, int_bt_curr_and_prev.days)
                dod = curr_date + datetime.timedelta(days=rnd_days_after_curr)

                value = {self.name: self.deceased_true_value}

                if self.deceased_date_field:
                    value[self.deceased_date_field] = dod.strftime(self.date_format)

                if self.age_field:
                    value[self.age_field] = curr_age

                return value

            prev_date = curr_date
            curr_date = curr_date + datetime.timedelta(weeks=52)

        value = {self.name: self.deceased_false_value}
        if self.deceased_date_field:
            value[self.deceased_date_field] = ""

        if self.age_field:
            value[self.age_field] = curr_age

        return value



@attr.s(kw_only=True)
class AgeField(Field):
    """
    Calculates age in years from two fields or values. The specified from_value and to_value can be either strings,
    date objects or Fields. If the former, then it is treated as the name of the field to obtain from the row.
    """
    from_value = attr.ib()
    to_value = attr.ib()
    from_format = attr.ib(default=None)
    to_format = attr.ib(default=None)

    def _next_value(self, row: Dict[str, Any]):
        from_date = extract_date(self.from_value, row, self.from_format)
        to_date = extract_date(self.to_value, row, self.to_format)
        return calculate_age(from_date, to_date)



from headfake.transformer import UpperCase, IntermittentBlanks
from headfake.field import ConstantField, LastNameField, GenderField
from headfake.fieldset import Fieldset


def test_upper_case_transformer():
    constant = ConstantField(value="my_constant", transformers = [UpperCase()])
    assert constant.next_value({}) == "MY_CONSTANT"

def test_multiple_transformers_works_within_multiple_field_fieldset():
    fs = Fieldset(fields = {
        "last_name": LastNameField(gender_field="gender", transformers = [
        IntermittentBlanks(blank_probability=0.2, blank_value=None), UpperCase()
        ]),
        "gender":GenderField(male_value=1,female_value=0)}
    )
    fs.field_map["last_name"].next_value({"gender":"M"})
